/*
 Small Programmable Power Supply
 Version: 0.5
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 8/09/2013
 Last Modified: 18/11/2013
 
 A small programmable power supply based on a the LM2596 module found on Ebay and digital pot to adjust the output 
 voltage with an ATtiny44/84. Uses a 4 digit LED segment display with 3 programmable buttons and a rotary encoder.
 
 Voltage input: 12-15V
 Voltage output: 1.8-12V
 
 */

// ATtiny24/44/84 Pin map
//
//                       +-\/-+
//                 VCC  1|o   |14  GND
//        Button 2 PB0  2|    |13  PA0 Rotary encoder A
//         Dpot CS PB1  3|    |12  PA1 Button 1
//           Reset PB3  4|    |11  PA2 ADC for Voltage
//     Shift Latch PB2  5|    |10  PA3 Rotary encoder B
//        Button 3 PA7  6|    |9   PA4 SCK
//        Button 4 PA6  7|    |8   PA5 MO
//                       +----+

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include "setup.c"

#define adcPin 2

byte readADC = 0; // When to read the ADC
int programButton = 0; // Which programmable button to select
volatile byte dpotValue = 255; // Digital pot value
volatile int dpotDisplaycounter = 0; // Used to show the dpot value for a little while
byte ledsBlink = 0; // Used for blinking LEDs
int blinkCounter = 0; // Blinking LEDs counter
bool calibratingDpot = false;
double previousReading = 0; // Previous ADC reading when calibrating
byte currentStep = 0; // Current step in the step index (when calibrated)

volatile byte numberDisplay[4] = {0, 0, 0, 0}; // Storage of the number to display
volatile byte segmentUpto = 0; // Which segment to light up
volatile byte buttonLeds = 0; // Which button LEDs to light up
volatile byte encoderWaitforlow = false;
volatile byte encoderTimeout = true;

// Write to the Dpot
void write_dpot(void) {
	dpot_latch_low;
	spi_transfer(0); // Command byte: Write data
	spi_transfer(dpotValue);
	dpot_latch_high;
}

// Shift out the numbers on the segment displays, no need to delay because the timer does that for us
void display_led(byte first, byte second, byte third, byte fourth) {
	if (segmentUpto == 0) {
		latch_low;
		spi_transfer(0x0E | buttonLeds); // First segment GND
		spi_transfer(number_to_led[first]);
		latch_high;
	}
	else if (segmentUpto == 1) {
		latch_low;
		spi_transfer(0x0D | buttonLeds); // Second segment GND
		spi_transfer(number_to_led[second] | (1<<NDP)); // Add dot poin
		latch_high;
	}
	else if (segmentUpto == 2) {
		latch_low;
		spi_transfer(0x0B | buttonLeds);
		spi_transfer(number_to_led[third]);
		latch_high;
	}
	else if (segmentUpto == 3) {
		latch_low;
		spi_transfer(0x07 | buttonLeds);
		spi_transfer(number_to_led[fourth]);
		latch_high;
		segmentUpto = 255; // Reset to zero on next increment
	}
	segmentUpto++;
}

// Store the number to be displayed on the screen
void store_number(double number) {
	if (number > 10) {
		numberDisplay[0] = (int) number / 10;
		numberDisplay[1] = (int) number % 10;
		numberDisplay[2] = (int) (number * 10) % 10;
		numberDisplay[3] = (int) (number * 100) % 10;
	}
	else {
		numberDisplay[0] = BLANK;
		numberDisplay[1] = (int) number;
		numberDisplay[2] = (int) (number * 10) % 10;
		numberDisplay[3] = (int) (number * 100) % 10;
	}
}

int main(void) {
	setup();
	
	while(1) {
		// Read voltage
		if (readADC == 0 && dpotDisplaycounter == 0) {
			int adcValue = analog_read_average(adcPin, REF1_1V); 
			double calculateDouble = (double) ((double) (adcValue) * 1.074) * 0.027015; // Calibrate this for higher accuracy
			store_number(calculateDouble);
			
			// Calibrate the digital pot so each step equates to a 0.1V step or close to it
			if (calibratingDpot == true) {
				// Check if reading minus previous reading is 0.1V difference
				if (dpotValue > 0 && ((calculateDouble - previousReading) > 0.1)) {
					buttonLeds = LED1 | LED2 | LED3;
					_delay_ms(250);
					previousReading = calculateDouble;
					stepIndex[currentStep][DPOTVAL] = dpotValue;
					stepIndex[currentStep][ADCVAL] = adcValue / 3;
					currentStep++;
				}
				else if (dpotValue == 0) { // No more to go, finished
					stepIndex[currentStep][DPOTVAL] = dpotValue; // Add top value
					stepIndex[currentStep][ADCVAL] = adcValue / 3;
					currentStep++;
					
					calibratingDpot = false;
					calibratedDpot = true;
					numberDisplay[0] = LETTER_D;
					numberDisplay[1] = LETTER_O;
					numberDisplay[2] = LETTER_N;
					numberDisplay[3] = LETTER_E;
					_delay_ms(2000);
					
					// Store in EEPROM
					eeprom_write_byte((uint8_t*) NO_OF_STEPS_LOCATION, currentStep);
					int y = STEPS_START_LOCATION;
					for (int x = 0; x < currentStep; x++) {
						eeprom_write_byte((uint8_t*) y++, stepIndex[x][DPOTVAL]);
						eeprom_write_byte((uint8_t*) y++, stepIndex[x][ADCVAL]);
					}
					currentStep = 0;
					
					// Clear programmed buttons
					for (int x = BUTTON_START_LOCATION; x <= BUTTON_END_LOCATION; x++) {
						eeprom_write_byte((uint8_t*) x, 0);
					}
				}
				else { // Increase voltage
					dpotValue--;
					buttonLeds = 0;
					_delay_ms(2); // Wait 
				}
			}
		}
		readADC++;
		
		if (calibratingDpot == false) {
			// First saved button 
			if (PINA & (1<<PROG_BUTTON_1)) {
				programButton = 0;
				if (calibratedDpot == true) { // Use steps
					currentStep = eeprom_read_byte((uint8_t*) programButton);
				}
				else {
					dpotValue = eeprom_read_byte((uint8_t*) programButton);
				}
				buttonLeds = LED1;
				_delay_ms(250); // Wait a little while for button to be released
			}
			// Second saved button 
			else if (PINB & (1<<PROG_BUTTON_2)) {
				programButton = 1;
				if (calibratedDpot == true) { // Use steps
					currentStep = eeprom_read_byte((uint8_t*) programButton);
				}
				else {
					dpotValue = eeprom_read_byte((uint8_t*) programButton);
				}
				buttonLeds = LED2;
				_delay_ms(250);
			}
			// Third saved button 
			else if (PINA & (1<<PROG_BUTTON_3)) {
				programButton = 2;
				if (calibratedDpot == true) { // Use steps
					currentStep = eeprom_read_byte((uint8_t*) programButton);
				}
				else {
					dpotValue = eeprom_read_byte((uint8_t*) programButton);
				}
				buttonLeds = LED3;
				_delay_ms(250);
			}
			// Set button
			else if (PINA & (1<<PROG_BUTTON_SET)) {
				bool buttonHeld = true;
				for (int x = 0; x < 20; x++) { // Check if held down for 1 second
					_delay_ms(50);
					if (! (PINA & (1<<PROG_BUTTON_SET))) {
						buttonHeld = false;
						break;
					}
				}
				
				// Menu
				if (buttonHeld == true) {
					byte menuOption = 0;
					bool menuOptionselected = false;
					buttonLeds = 0;
					dpotValue = 255;
					
					while (1) { // Wait until the set button is pressed
						if (menuOption == 0) { // Calibrate
							numberDisplay[0] = LETTER_C;
							numberDisplay[1] = LETTER_A;
							numberDisplay[2] = LETTER_L;
							numberDisplay[3] = BLANK;
							
							if (menuOptionselected == true) {
								calibratingDpot = true;
								buttonLeds = 0;
								previousReading = 0;
								currentStep = 0;
								buttonLeds = LED1 | LED2 | LED3;
								_delay_ms(5000); // Wait for dpot to drop to the lowest setting
								break;
							}
						}
						else if (menuOption == 1) { // Reset calibration
							numberDisplay[0] = LETTER_R;
							numberDisplay[1] = LETTER_S;
							numberDisplay[2] = LETTER_E;
							numberDisplay[3] = LETTER_T;
							
							if (menuOptionselected == true) {
								calibratedDpot = false;
								
								// Clear steps and programmed buttons
								eeprom_write_byte((uint8_t*) NO_OF_STEPS_LOCATION, 255);
								for (int x = BUTTON_START_LOCATION; x <= BUTTON_END_LOCATION; x++) {
									eeprom_write_byte((uint8_t*) x, 0);
								}
								
								break;
							}
						}
						
						// Button presses
						if (PINA & (1<<PROG_BUTTON_1)) {
							menuOption = 0;
							_delay_ms(250);
						}
						else if (PINB & (1<<PROG_BUTTON_2)) {
							menuOption = 1;
							_delay_ms(250);
						}
						else if (PINA & (1<<PROG_BUTTON_3)) {
							_delay_ms(250);
							break;
						}
						else if (PINA & (1<<PROG_BUTTON_SET)) {
							if (buttonHeld == false) { // Wait until user lets go of set button initally
								menuOptionselected = true;
							}
							_delay_ms(250);
						}
						else {
							buttonHeld = false;
						}
					}
				}
				else {
					blinkCounter = 1000; // Blink LED when saving setting
					ledsBlink = buttonLeds;
					
					if (calibratedDpot == true) {
						eeprom_write_byte((uint8_t*) programButton, currentStep);
					}
					else {
						eeprom_write_byte((uint8_t*) programButton, dpotValue);
					}
				}
			}
			
			// If Dpot calibrated, then use the steps
			if (calibratedDpot == true) {
				dpotValue = stepIndex[currentStep][DPOTVAL];
			}
		}
		
		// Debug: Display digital pot value
		//numberDisplay[1] = (dpotValue/100);
		//numberDisplay[2] = ((dpotValue/10) %10);
		//numberDisplay[3] = (dpotValue % 10);
		
		// Used for timings
		if (dpotDisplaycounter > 0) {
			dpotDisplaycounter--;
		}
		if (blinkCounter > 0) {
			if (blinkCounter % 128 == 0) { // Blink LED
				buttonLeds ^= ledsBlink;
			}
			blinkCounter--;
			
			if (blinkCounter == 0) { // Leave LED lit when finished
				buttonLeds |= ledsBlink;
			}
		}
		_delay_us(500);
		
	}
}

ISR(ADC_vect) { }

// Timer0 overflow after 2ms, used to update LED and Dpot
ISR(TIM0_OVF_vect) {
	display_led(numberDisplay[0], numberDisplay[1], numberDisplay[2], numberDisplay[3]);
	write_dpot();
}

// Timer1 compare
ISR(TIM1_COMPA_vect) {
	cbi(TCCR1B, CS10);
	encoderTimeout = true;
}

// Pin interrupt on rotary encoder pins
ISR(PCINT0_vect) {
	// Waiting for either pin to go low
	if (encoderWaitforlow == true && !((PINA & (1<<PA3)) && (PINA & (1<<PA0)))) {
		if (calibratedDpot == true) { // Use steps
			if ((PINA & (1<<PA3)) && currentStep < 50) { // See which pin is low
				currentStep++;
			}
			else if ((PINA & (1<<PA0)) && currentStep > 0) {
				currentStep--;
			}
			
			// Display calculated value
			dpotDisplaycounter = 400;
			double dpotTemp = (double) ((double) (stepIndex[currentStep][ADCVAL] * 3) * 1.074) * 0.027015;
			store_number(dpotTemp);
		}
		else { // Don't use steps
			if ((PINA & (1<<PA3)) && dpotValue > 0) { // See which pin is low
				dpotValue--;
			}
			else if ((PINA & (1<<PA0)) && dpotValue < 255) {
				dpotValue++;
			}
		}
		
		// Wait for 3ms
		OCR1A = 3000;
		TCNT1 = 0;
		sbi(TCCR1B, CS10);
		encoderWaitforlow = false;
		encoderTimeout = false;
	}
	else if (encoderWaitforlow == false && encoderTimeout == true) { // After the 3ms
		if (!((PINA & (1<<PA3)) && (PINA & (1<<PA0)))) { // If either pin is low, wait another 500uS
			OCR1A = 500;
			TCNT1 = 0;
			sbi(TCCR1B, CS10);
		}
		else {
			encoderWaitforlow = true;
		}
	}
}
